package naroth.sound;

import java.io.*;
import java.nio.*;
import javax.sound.sampled.*;

import org.lwjgl.*;
import org.lwjgl.openal.*;
import com.threed.jpct.*;


public class OpenALSound implements Sound{

   private int buffer=-1;
   private boolean ok=false;

   private int lastIndex=-1;
   private int currentLoopIndex=-1;

   private static IntBuffer scratchBuffer = BufferUtils.createIntBuffer(256);
   private static int[] sources;
   private static int[] buffers=null;
   private static int sourceIndex;
   private static int channels=16;
   private static int instanceCount=0;
   private boolean mute=false;

   static {
      try {

         boolean ok=false;
         do {
            AL.create();
            sources=new int[channels];
            Logger.log("Trying with "+channels+" channels!", Logger.MESSAGE);
            scratchBuffer.limit(channels);
            AL10.alGenSources(scratchBuffer);
            if (AL10.alGetError()==AL10.AL_NO_ERROR) {
               scratchBuffer.rewind();
               scratchBuffer.get(sources);
               ok=true;
            } else {
               //AL10.alDeleteSources(scratchBuffer);
               AL.destroy();
               channels=channels>>1;
            }
         } while (!ok&&channels>0);

         if (channels!=0) {
            Logger.log("OpenAL initialized using "+channels+" channels!", Logger.MESSAGE);
         } else {
            Logger.log("Can't initialize one single channel!?", Logger.WARNING);
         }

      } catch (Exception e) {
         Logger.log("Unable to initialize OpenAL!", Logger.WARNING);
      }
   }

   public OpenALSound(InputStream is) {
      try {

         instanceCount++;
         IntBuffer scratchBuffer=BufferUtils.createIntBuffer(256);
         scratchBuffer.rewind().position(0).limit(1);
         AL10.alGenBuffers(scratchBuffer);
         buffer=scratchBuffer.get(0);

         byte[] data=null;

         Logger.log("Loading sound (OpenAL) from InputStream", Logger.MESSAGE);
         AudioInputStream soundStream=AudioSystem.getAudioInputStream(is);
         AudioFormat format=soundStream.getFormat();

         int len=(int) (format.getFrameSize()*soundStream.getFrameLength());
         data=new byte[len];
         int l=0;
         int lp=0;
         do {
            l=soundStream.read(data, lp, len);
            lp+=l;
         } while (l!=-1);

         // Welches Format?
         int chans = 0;
         boolean bit16=false;

         if (format.getChannels()==1) {
            if (format.getSampleSizeInBits()==8) {
               chans=AL10.AL_FORMAT_MONO8;
            } else if (format.getSampleSizeInBits()==16) {
               chans=AL10.AL_FORMAT_MONO16;
               bit16=true;
            }
         } else if (format.getChannels()==2) {
            if (format.getSampleSizeInBits()==8) {
               chans=AL10.AL_FORMAT_STEREO8;
            } else if (format.getSampleSizeInBits()==16) {
               chans=AL10.AL_FORMAT_STEREO16;
               bit16=true;
            }
         }

         ByteBuffer bb = ByteBuffer.allocateDirect(data.length);
         bb.order(ByteOrder.nativeOrder());
         ByteBuffer src=ByteBuffer.wrap(data);
         src.order(ByteOrder.LITTLE_ENDIAN);
         if (bit16) {
            ShortBuffer destShort=bb.asShortBuffer();
            ShortBuffer srcShort=src.asShortBuffer();
            while (srcShort.hasRemaining())
               destShort.put(srcShort.get());
         } else {
            while (src.hasRemaining())
               bb.put(src.get());
         }
         bb.rewind();

         AL10.alBufferData(buffer, chans, bb, (int) format.getSampleRate());

      } catch (Exception e) {
         Logger.log("Error loading sound: "+e, Logger.ERROR);
      }
   }

   public boolean playedFine() {
      return ok;
   }

   public boolean isRunning() {
      return lastIndex!=-1 && AL10.alGetSourcei(sources[lastIndex], AL10.AL_SOURCE_STATE) == AL10.AL_PLAYING;
   }

   public void stop() {
      if (isRunning()) {
         AL10.alSourceStop(sources[lastIndex]);
         lastIndex=-1;
      }
   }

   public void setMute(boolean is) {
      mute=is;
   }

   public void play() {
      if (mute) { return; }
      int channel=getNextChannel();
      if (channel!=-1) {
         AL10.alSourcei(sources[channel], AL10.AL_BUFFER, buffer);
         AL10.alSourcePlay(sources[channel]);
      }
   }

   public void endLoop() {
      if (currentLoopIndex!=-1) {
         AL10.alSourcei(sources[currentLoopIndex], AL10.AL_LOOPING, 0);
         currentLoopIndex=-1;
      }
   }

   public void loop() {
      if (mute) { return; }
      if (currentLoopIndex==-1) {
         currentLoopIndex=getNextChannel();
         if (currentLoopIndex!=-1) {
            AL10.alSourcei(sources[currentLoopIndex], AL10.AL_BUFFER, buffer);
            AL10.alSourcei(sources[currentLoopIndex], AL10.AL_LOOPING, 1);
            AL10.alSourcePlay(sources[currentLoopIndex]);
         }
      }
   }

   private int getNextChannel() {
      int cnt=0;
      boolean ok=false;
      while (cnt<channels && !ok) {
         sourceIndex=(sourceIndex+1)%sources.length;
         cnt++;
         ok=AL10.alGetSourcei(sources[sourceIndex], AL10.AL_SOURCE_STATE) != AL10.AL_PLAYING;
      }

      if (cnt>=channels) {
         ok=false;
         return -1;
      }
      ok=true;
      lastIndex=sourceIndex;
      return sourceIndex;
   }


   public synchronized void destroy() {

      if (buffers==null) {
         buffers=new int[instanceCount];
      }
      // Einsammeln der Buffer...
      buffers[instanceCount-1]=buffer;

      instanceCount--;
      if (instanceCount==0) {

         // Absolute Grtze, aber aufgrund der static-Eigenschaft der OpenAL-Anbindung
         // leider nicht anders zu machen.

         Logger.log("Destroying OpenAL!", Logger.MESSAGE);

         scratchBuffer.position(0).limit(sources.length);
         scratchBuffer.put(sources).flip();
         AL10.alSourceStop(scratchBuffer);

         AL10.alDeleteSources(scratchBuffer);
         sources=null;

         // Die Buffer selber lschen
         for (int i=0; i<buffers.length; i++) {
            scratchBuffer.position(0).limit(1);
            scratchBuffer.put(buffers[i]).flip();
            AL10.alDeleteBuffers(scratchBuffer);
         }

         buffers=null;
         AL.destroy();
      }
   }
}
